@@ -1,6 +1,5 @@
-/* Copyright (C) 1995, 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003,
- *   2004, 2006, 2007, 2008, 2009, 2010, 2011, 2012, 2013,
- *   2014, 2015, 2017 Free Software Foundation, Inc.
+/* Copyright (C) 1995-2004, 2006-2015, 2017, 2019
+ *   Free Software Foundation, Inc.
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public License
@@ -614,9 +613,9 @@ static scm_t_off
 fport_seek (SCM port, scm_t_off offset, int whence)
 {
   scm_t_fport *fp = SCM_FSTREAM (port);
-  off_t_or_off64_t result;
+  scm_t_off result;
 
-  result = lseek_or_lseek64 (fp->fdes, offset, whence);
+  result = lseek (fp->fdes, offset, whence);
 
   if (result == -1)
     scm_syserror ("fport_seek");
